package org.jvnet.annox.model;

import java.lang.annotation.Annotation;
import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.lang.reflect.Proxy;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class XAnnotation {

	private final Class<? extends Annotation> annotationClass;

	private final XAnnotationField<?>[] fields;

	private final Map<String, XAnnotationField<?>> fieldsMap;

	public XAnnotation(final Class<? extends Annotation> annotationClass,
			final XAnnotationField<?>[] fields) throws IllegalArgumentException {
		if (annotationClass == null)
			throw new IllegalArgumentException(
					"Annotation class must not be null.");
		this.annotationClass = annotationClass;
		if (fields == null) {
			this.fields = new XAnnotationField[0];
			this.fieldsMap = Collections.emptyMap();
		} else {
			final Map<String, XAnnotationField<?>> fieldsMap = new HashMap<String, XAnnotationField<?>>();
			for (XAnnotationField<?> field : fields) {
				if (field == null)
					throw new IllegalArgumentException(
							"Fields array must not contain null entries.");
				fieldsMap.put(field.getName(), field);
			}
			this.fields = fields;
			this.fieldsMap = Collections.unmodifiableMap(fieldsMap);
		}
		check();
	}

	private void check() {

		final Method[] methods = annotationClass.getMethods();

		final List<XAnnotationField<?>> fields = new ArrayList<XAnnotationField<?>>(
				methods.length);

		for (final Method method : methods) {
			if (!Annotation.class.equals(method.getDeclaringClass())) {
				final String name = method.getName();
				final Class<?> type = method.getReturnType();
				final XAnnotationField<?> field = fieldsMap.get(name);
				if (field == null) {
					throw new IllegalArgumentException("Field [" + name
							+ "] is not defined.");
				}
				final Class<?> fieldType = field.getType();
				if (!type.equals(fieldType)) {
					throw new IllegalArgumentException("Field [" + name
							+ "] has the wrong type [" + field.getType()
							+ "] instead of expected [" + type + "].");
				}
			}
		}

	}

	public Class<? extends Annotation> getAnnotationClass() {
		return annotationClass;
	}

	public XAnnotationField<?>[] getFields() {
		return fields;
	}

	@Override
	public int hashCode() {
		int hash = 0;
		hash = hash * 37 + annotationClass.hashCode();
		for (int i = 0; i < fields.length; i++) {
			hash = hash * 37 + fields.hashCode();
		}
		return hash;
	}

	@Override
	public boolean equals(Object obj) {
		if (!(obj instanceof XAnnotation)) {
			return false;
		}
		if (this == obj) {
			return true;
		}
		final XAnnotation other = (XAnnotation) obj;
		if (annotationClass.equals(other.annotationClass)) {
			final XAnnotationField[] lhs = fields;
			final XAnnotationField[] rhs = other.fields;
			if (lhs == rhs) {
				return true;
			}
			if (lhs == null || rhs == null) {
				return false;
			}
			if (lhs.length != rhs.length) {
				return false;
			}
			for (int i = 0; i < lhs.length; ++i) {
				if (!(lhs[i].equals(rhs[i])))
					return false;
			}
			return true;
		} else
			return false;
	}

	@Override
	public String toString() {
		final StringBuffer sb = new StringBuffer();
		sb.append("@");
		sb.append(getAnnotationClass().getName());

		sb.append("(");
		for (int index = 0; index < fields.length; index++) {
			final XAnnotationField<?> field = fields[index];
			if (index > 0) {
				sb.append(", ");
			}
			sb.append(field);
		}
		sb.append(")");
		return sb.toString();
	}

	public Annotation getResult() {
		final Map<String, Object> draft = new HashMap<String, Object>();

		for (XAnnotationField<?> field : fields) {
			if (field == null)
				throw new IllegalArgumentException(
						"Fields array must not contain null entries.");
			draft.put(field.getName(), field.getResult());
		}
		final Map<String, Object> map = Collections.unmodifiableMap(draft);

		final InvocationHandler handler = new InvocationHandler() {

			public Object invoke(Object proxy, Method method, Object[] args)
					throws Throwable {

				final String name = method.getName();
				if (!Annotation.class.equals(method.getDeclaringClass())) {
					return map.get(name);
				} else {
					if ("annotationType".equals(name)) {
						return annotationClass;
					} else {
						return null;
					}
				}
			}
		};

		final Annotation annotation = (Annotation) Proxy.newProxyInstance(
				annotationClass.getClassLoader(),
				new Class[] { annotationClass }, handler);

		return annotation;
	}

	public <T> T accept(XAnnotationVisitor<T> visitor) {
		return visitor.visitAnnotation(this);
	}
}
