package org.jvnet.annox.model;

import java.lang.annotation.Annotation;
import java.lang.reflect.Array;

import org.apache.commons.lang.Validate;
import org.jvnet.annox.util.ArrayUtils;

public abstract class XAnnotationField<T> {

	private final String name;

	public String getName() {
		return name;
	}

	private final T value;

	public T getValue() {
		return value;
	}

	public abstract Object getResult();

	public abstract Class<?> getType();

	public abstract <T> T accept(XAnnotationVisitor<T> visitor);

	public XAnnotationField(final String name, final T value) {
		Validate.notNull(name, "Field name must not be null.");
		Validate.notNull(value, "Field value must not be null.");
		this.name = name;
		this.value = value;
	}

	public static abstract class XField<T> extends XAnnotationField<T> {
		public XField(String name, T value) {
			super(name, value);
		}

		@Override
		public String toString() {
			return getName() + "=" + getValue();
		}

		@Override
		public int hashCode() {
			int hash = 0;
			final String name = getName();
			hash = hash * 37 + name.hashCode();
			final T value = getValue();
			hash = hash * 37 + value.hashCode();
			return hash;
		}

		@Override
		public boolean equals(Object obj) {
			if (!(obj instanceof XAnnotationField)) {
				return false;
			}
			if (this == obj) {
				return true;
			}
			final XAnnotationField other = (XAnnotationField) obj;
			return getName().equals(other.getName())
					&& getValue().equals(other.getValue());
		}

		@Override
		public Object getResult() {
			return getValue();
		}
	}

	public static abstract class XPrimitiveField<T> extends XField<T> {
		public XPrimitiveField(String name, T value) {
			super(name, value);
		}
	}

	public static abstract class XObjectField<T> extends XField<T> {
		public XObjectField(String name, T value) {
			super(name, value);
		}
	}

	public static class XBoolean extends XPrimitiveField<Boolean> {
		public XBoolean(String name, Boolean value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Boolean.TYPE;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitBooleanField(this);
		}
	}

	public static class XByte extends XPrimitiveField<Byte> {
		public XByte(String name, Byte value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Byte.TYPE;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitByteField(this);
		}
	}

	public static class XInt extends XPrimitiveField<Integer> {
		public XInt(String name, Integer value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Integer.TYPE;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitIntField(this);
		}
	}

	public static class XLong extends XPrimitiveField<Long> {
		public XLong(String name, Long value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Long.TYPE;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitLongField(this);
		}
	}

	public static class XShort extends XPrimitiveField<Short> {
		public XShort(String name, Short value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Short.TYPE;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitShortField(this);
		}
	}

	public static class XChar extends XPrimitiveField<Character> {
		public XChar(String name, Character value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Character.TYPE;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitCharField(this);
		}
	}

	public static class XClass extends XObjectField<Class> {
		public XClass(String name, Class value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Class.class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitClassField(this);
		}
	}

	public static class XDouble extends XPrimitiveField<Double> {
		public XDouble(String name, Double value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Double.TYPE;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitDoubleField(this);
		}
	}

	public static class XFloat extends XPrimitiveField<Float> {
		public XFloat(String name, Float value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Float.TYPE;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitFloatField(this);
		}
	}

	public static class XEnum extends XObjectField<Enum> {

		private final Class<? extends Enum> enumClass;

		public XEnum(String name, Enum value, Class<? extends Enum> enumClass) {
			super(name, value);
			Validate.notNull(enumClass, "Enum class must not be null.");
			this.enumClass = enumClass;
		}

		@Override
		public Class<?> getType() {
			return enumClass;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitEnumField(this);
		}
	}

	public static class XString extends XObjectField<String> {
		public XString(String name, String value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return String.class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitStringField(this);
		}
	}

	public static class XAnnotation extends
			XObjectField<org.jvnet.annox.model.XAnnotation> {

		private final Class<? extends Annotation> annotationClass;

		public XAnnotation(String name,
				org.jvnet.annox.model.XAnnotation value,
				Class<? extends Annotation> annotationClass) {
			super(name, value);
			Validate.notNull(annotationClass,
					"Annotation class must not be null.");
			this.annotationClass = annotationClass;
		}

		@Override
		public Class<?> getType() {
			return annotationClass;
		}

		public Class<? extends Annotation> getAnnotationClass() {
			return annotationClass;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitAnnotationField(this);
		}

		@Override
		public Object getResult() {
			return getValue().getResult();
		}
	}

	public static abstract class XArrayField<T> extends XAnnotationField<T[]> {
		public XArrayField(String name, T[] value) {
			super(name, value);
			Validate.noNullElements(value,
					"Array must not contain null values.");
		}

		@Override
		public String toString() {
			final StringBuffer sb = new StringBuffer();
			sb.append(getName());
			sb.append("=[");
			final T[] values = getValue();
			for (int index = 0; index < values.length; index++) {
				final T value = values[index];
				if (index > 0) {
					sb.append(", ");
				}
				sb.append(value);
			}
			sb.append("]");
			return sb.toString();
		}

		@Override
		public int hashCode() {
			int hash = 0;
			final String name = getName();
			hash = hash * 37 + name.hashCode();
			final T[] values = getValue();
			for (int index = 0; index < values.length; index++) {
				final T v = values[index];
				hash = hash * 37 + v.hashCode();
			}
			return hash;
		}

		@Override
		public boolean equals(Object obj) {
			if (!(obj instanceof XArrayField<?>)) {
				return false;
			}
			if (this == obj) {
				return true;
			}
			final XArrayField<?> other = (XArrayField<?>) obj;
			if (!getName().equals(other.getName()))
				return false;

			final Object[] lhs = getValue();
			final Object[] rhs = other.getValue();
			if (lhs == rhs) {
				return true;
			}
			if (lhs == null || rhs == null) {
				return false;
			}
			if (lhs.length != rhs.length) {
				return false;
			}
			for (int i = 0; i < lhs.length; ++i) {
				if (!(lhs[i].equals(rhs[i])))
					return false;
			}
			return true;
		}

		@Override
		public Object getResult() {
			return ArrayUtils.asPrimitiveArray(getValue());
		}

	}

	public static abstract class XPrimitiveArrayField<T> extends XArrayField<T> {
		public XPrimitiveArrayField(String name, T[] values) {
			super(name, values);
		}
	}

	public static abstract class XObjectArrayField<T> extends XArrayField<T> {
		public XObjectArrayField(String name, T[] values) {
			super(name, values);
		}
	}

	public static class XBooleanArray extends XPrimitiveArrayField<Boolean> {
		public XBooleanArray(String name, Boolean[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return boolean[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitBooleanArrayField(this);
		}
	}

	public static class XByteArray extends XPrimitiveArrayField<Byte> {
		public XByteArray(String name, Byte[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return byte[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitByteArrayField(this);
		}
	}

	public static class XIntArray extends XPrimitiveArrayField<Integer> {
		public XIntArray(String name, Integer[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return int[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitIntArrayField(this);
		}
	}

	public static class XLongArray extends XPrimitiveArrayField<Long> {
		public XLongArray(String name, Long[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return long[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitLongArrayField(this);
		}
	}

	public static class XShortArray extends XPrimitiveArrayField<Short> {
		public XShortArray(String name, Short[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return short[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitShortArrayField(this);
		}
	}

	public static class XCharArray extends XPrimitiveArrayField<Character> {
		public XCharArray(String name, Character[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return char[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitCharArrayField(this);
		}
	}

	public static class XClassArray extends XObjectArrayField<Class> {
		public XClassArray(String name, Class[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return Class[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitClassArrayField(this);
		}
	}

	public static class XDoubleArray extends XPrimitiveArrayField<Double> {
		public XDoubleArray(String name, Double[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return double[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitDoubleArrayField(this);
		}

	}

	public static class XFloatArray extends XPrimitiveArrayField<Float> {
		public XFloatArray(String name, Float[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return float[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitFloatArrayField(this);
		}
	}

	public static class XEnumArray extends XObjectArrayField<Enum> {

		private final Class<? extends Enum> enumClass;

		public XEnumArray(String name, Enum[] value,
				Class<? extends Enum> enumClass) {
			super(name, value);
			Validate.notNull(enumClass, "Enum class must not be null.");
			this.enumClass = enumClass;
		}

		@Override
		public Class<?> getType() {
			return Array.newInstance(enumClass, 0).getClass();
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitEnumArrayField(this);
		}
	}

	public static class XStringArray extends XObjectArrayField<String> {
		public XStringArray(String name, String[] value) {
			super(name, value);
		}

		@Override
		public Class<?> getType() {
			return String[].class;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitStringArrayField(this);
		}
	}

	public static class XAnnotationArray extends
			XObjectArrayField<org.jvnet.annox.model.XAnnotation> {

		private final Class<? extends Annotation> annotationClass;

		public XAnnotationArray(String name,
				org.jvnet.annox.model.XAnnotation[] value,
				Class<? extends Annotation> annotationClass) {
			super(name, value);
			Validate.notNull(annotationClass,
					"Annotation class must not be null.");
			this.annotationClass = annotationClass;
		}

		@Override
		public Class<?> getType() {
			return Array.newInstance(annotationClass, 0).getClass();
		}

		public Class<? extends Annotation> getAnnotationClass() {
			return annotationClass;
		}

		@Override
		public <T> T accept(XAnnotationVisitor<T> visitor) {
			return visitor.visitAnnotationArrayField(this);
		}

		@Override
		public Object getResult() {
			final org.jvnet.annox.model.XAnnotation[] value = getValue();
			final Annotation[] result = (Annotation[]) Array.newInstance(
					annotationClass, value.length);
			for (int index = 0; index < value.length; index++) {
				final org.jvnet.annox.model.XAnnotation annotation = value[index];
				result[index] = (Annotation) annotation.getResult();
			}
			return result;
		}
	}
}
